#
# main.py
#

import RobotCanvas
import Tkinter
import time
import threading
# the python Queue is a data structure that is
# thread-safe; we use it to pass messages
# between the visualizer and the main loop
import Queue    
import random
import math
import sys
from ThreadHandler import *
from copy import deepcopy

        
class MainLoop(ThreadHandler):
    """
    A class defining the sense-plan-act "main" loop, named main
    """
    
    def __init__(self, mapname):
        """
        This constructor delegates a couple graphics-specific things
        to the base class, MainClientThread
        """
        self.mapname = mapname
        ThreadHandler.__init__(self, mapname)
        #
        # here would be a good place to read in data from
        # configuration files...
        #
        
        
    def main(self):
        """
        This while loop is, in essence, the life of the robot.
        Each time through the loop, you will want to 
           * read the sensors (SENSE)
           * decide on what action to take (PLAN)
           * execute the current action (ACT)
        """
        # pause for a second to let the graphics initialize
        time.sleep(1)

        # these constants are the robot's translational
        # and rotational velocities...
        FORWARD_VELOCITY = 50.0   # 50.0 cm/sec is the max
        FV = FORWARD_VELOCITY     # short names are good
        ROTATIONAL_VELOCITY = 40.0    # 75.0 deg/sec is the max
        RV = ROTATIONAL_VELOCITY
        NUMPARTICLES=50

        # here is the map
        segList = self.getMap()
        counter = 0
        for s in segList:
            print 'segment #', counter, 'is (', s.x1, ',', s.y1, ') - (', s.x2, ',', s.y2, ')'
            counter += 1
        
        noise = 0.05         # noise in the odometry

        particles = []       # the list of particles
        lastx, lasty, lastthr = 0, 0, 0
        
        while True:          # run until you hit 'Q' in the window 
                             # or control-c at the command prompt
            
            # SENSE
            #
            # x is the x coordinate of the robot
            # y is the y coordinate of the robot
            # thd is the robot's heading, theta, in degrees
            # bump is a list of two bump sensor readings: left and right
            # dist is the straight-ahead range to the nearest wall
            realpose, odometry, bump, dist, color = self.getData()

            gx, gy, gthr = odometry    # unpack the odometry (in global coordinates)
            
            # find the delta x, y, and theta since the last update
            dtr = gthr - lastthr
            gdx = gx - lastx
            gdy = gy - lasty
            
            # find the dx and dy in local robot coordinates
            dx =  gdx*math.cos(gthr) + gdy*math.sin(gthr)
            dy = -gdx*math.sin(gthr) + gdy*math.cos(gthr)

            # GUI - get any keypresses and goal locations (mouse clicks)
            #
            kbd, goal = self.getWindowEvents()  # at start, kbd == '' and goal == (0,0)

            if   kbd == 'i': self.setVels(FV,0,noise)
            elif kbd == 'k': self.setVels(-FV,0,noise)
            elif kbd == 'j': self.setVels(FV,RV,noise)
            elif kbd == 'l': self.setVels(FV,-RV,noise)
            elif kbd == ' ': # space stops and retains kbd control
                self.setVels(0,0)
            elif kbd == 'P':
                realx, realy, realthd = realpose
                print "realpose is", realpose
                realthr = math.radians(realthd)
                particles = []
                for i in range(NUMPARTICLES):  # 10 particles for now
                    rr = abs(random.gauss(0,20))#20
                    rtheta = random.uniform(-math.pi, math.pi)
                    rthetaor = random.gauss(0, math.pi/16)+realthr #16
                    newx = realx+rr*math.cos(rtheta)
                    newy = realy+rr*math.sin(rtheta)
                    print "  particle", i, " pose is ", (newx, newy, rthetaor),
                    data = self.getRangeData( newx, newy, rthetaor )
                    print " with data=", data
                    particles.append([newx, newy, rthetaor, 1.0, data[0]])
                    # the color is in data, too - it's just not being used yet
                    # this is the drawing command...
                self.queue.put([self.setParticles, particles])                    
            # special cases to allow us to quit entirely ('Q')
            # or to stop (' ') and resume keyboard control
            elif kbd == 'Q': break    # 'Q' quits the program


            ## You need to modify the motion and sensor model functions
            if abs(dtr) > 0 or abs(dx) > 0 or abs(dy) > 0:
                self.apply_motion_model(particles, dx, dy, dtr)
                self.apply_sensor_model(particles, realpose)
                self.normalize_particles(particles)                
                self.resample_particles(particles)
                self.queue.put([self.setParticles, particles])                    
                lastx, lasty, lastthr = gx, gy, gthr

            # We sleep to let the graphics catch up...
            # Don't sleep elsewhere (it makes the program
            # unresponsive, since the kbd is not being
            # checked!) Instead, use the PAUSE state.
            time.sleep(0.05) # 1/40 of a second
            #time.sleep(1)
            
        # if the while loop ends, we shutdown the gui, too.
        print "Main loop stopped."
        self.shutdown()
    
            
    def apply_motion_model(self, particles, dx, dy, dt):
        '''Move the particles according to the motion model of the robot'''

        for p in particles:
            
            # convert the local coordinates back into global coordinates
            gdx = dx*math.cos(p[2]) - dy*math.sin(p[2])
            gdy = dx*math.sin(p[2]) + dy*math.cos(p[2])

            p[0] += gdx
            p[1] += gdy
            p[2] += dt


    def apply_sensor_model(self, particles, realpose):
        '''Use the range sensors to set the weight of the particles'''
        
        # get sensor measurement from range sensor
        # orientation is in DEGREES in realpose
        rd, rr, rg, rb = self.getRangeData(realpose[0], 
                                           realpose[1], 
                                           math.radians(realpose[2]))            

        return


    def normalize_particles(self, particles):
        '''normalize the particles so the weights sum to one'''
        sum = 0.0
        for p in particles:
            sum += p[3]

        for p in particles:
            p[3] /= sum


    def resample_particles(self, particles):
        ''' resample the particles according to their weight values'''

        #sort the particles by their weights
        particles.sort(cmp_weights)

        # make a copy of the particles
        oldparticles = deepcopy(particles)

        #  resample from the set of old particles
        for i in range(len(particles)):
            r = random.random()
            mass = 0.0
            j = 0
            while j < len(oldparticles) and r > mass:
                mass += oldparticles[j][3]
                j += 1

            j -= 1
            particles[i] = deepcopy(oldparticles[j])

        # reset all the particles to have uniform weights 
        for p in particles:
            p[3] = 1.0 / len(particles)
         

def cmp_weights(p1, p2):
    if p1[3] < p2[3]:
        return 1
    elif p1[3] > p2[3]:
        return -1
    else:
        return 0
    
if __name__ == '__main__':
        
    # you might try one of these maps:
    MAPS = [ 'map0.txt',  # 0
             'map1.txt',  # 1
             'map2rooms.txt',   # 2
             'mapColors.txt',   # 3
             'mapMaze.txt',     # 4
             'mapHallways.txt', # 5
             'maphw1.txt' ]  # 6

    # creates a MainLoop object and then
    # calls its run method, which calls the
    # graphics system's mainloop...
    MainLoop(MAPS[6]).run()
    # I think it's important that run(), above,
    # is called right after the constructor so that
    # the graphical calls made by the main() method here
    # actually have a graphical framework to address, as
    # created by run. The 1-second delay at the start
    # of main helps ensure that things happen in the
    # right order.


